#include "spells.h"
#include <QDebug>
#include "interface.h"
#include "game.h"

CSpells spells;

//---------------------------------------------------------------------------------------------
CSpells::CSpells()
{

}

//---------------------------------------------------------------------------------------------
void CSpells::cast(int champion)
{
    CCharacter* c = &game.characters[champion];
    char*   spell = c->spell;
    int     level = spell[0] - 'a';

    // magic torch
    if (strcmp(&spell[1], "j") == 0)
    {
        CSpell  s;
        s.spell = eSpell_MagicTorch;
        s.power = (level + 4) * 16;
        s.decreaseTime = (level + 4) * 60;
        s.time.set(41250 + level * 8000 + s.decreaseTime, true);
        mActiveSpells.push_back(s);
        return;
    }

    // light
    else if (strcmp(&spell[1], "ipw") == 0)
    {
        CSpell  s;
        s.spell = eSpell_MagicTorch;
        s.power = (level * 2 + 4) * 16;
        s.decreaseTime = (level * 2 + 4) * 60;
        s.time.set(156250 + level * 32000 + s.decreaseTime, true);
        mActiveSpells.push_back(s);
        return;
    }

    // darkness
    else if (strcmp(&spell[1], "kpx") == 0)
    {
        CSpell  s;
        s.spell = eSpell_MagicTorch;
        s.power = -(level + 3) * 16;
        s.decreaseTime = (level + 3) * 60;
        s.time.set(1531 + s.decreaseTime, true);
        mActiveSpells.push_back(s);
        return;
    }

    std::string message = interface.setChampionNameInString(champion, "MESSAGE01");
    interface.addMessage(message, 4);
}

//---------------------------------------------------------------------------------------------
void CSpells::update()
{
    std::vector<CSpell>::iterator   it;

    for (it = mActiveSpells.begin(); it!= mActiveSpells.end();)
    {
        if (it->time.update() == true)
            it = mActiveSpells.erase(it);
        else
            it++;
    }
}

//---------------------------------------------------------------------------------------------
int CSpells::getMana(int champion, char symbol)
{
    CCharacter* c = &game.characters[champion];
    char*   spell = c->spell;

    if (spell[0] == 0)
    {
        return symbol - 'a' + 1;
    }
    else
    {
        int     level = spell[0] - 'a';
        static const int    costs[] =
            {
                2,  // g = Ya
                3,  // h = Vi
                4,  // i = Oh
                5,  // j = Ful
                6,  // k = Des
                7,  // l = Zo
                4,  // m = Ven
                5,  // n = Ew
                6,  // o = Kath
                7,  // p = Ir
                7,  // q = Bro
                9,  // r = Gor
                2,  // s = Ku
                2,  // t = Ros
                3,  // u = Dain
                4,  // v = Neta
                6,  // w = Ra
                7   // x = Sar
            };

        return (costs[symbol - 'g'] * (level + 2)) / 2;
    }
}

//---------------------------------------------------------------------------------------------
void CSpells::save(FILE *handle)
{
    int size = mActiveSpells.size();
    fwrite(&size, sizeof(size), 1, handle);

    for (int i = 0; i < size; ++i)
        mActiveSpells[i].save(handle);
}

//---------------------------------------------------------------------------------------------
void CSpells::load(FILE *handle)
{
    int size;
    fread(&size, sizeof(size), 1, handle);

    mActiveSpells.clear();
    mActiveSpells.resize(size);
    for (int i = 0; i < size; ++i)
        mActiveSpells[i].load(handle);
}

//---------------------------------------------------------------------------------------------
int CSpell::getPower()
{
    if (time.get() >= decreaseTime)
        return power;
    else
        return (power * time.get()) / decreaseTime;
}

//---------------------------------------------------------------------------------------------
void CSpell::save(FILE *handle)
{
    fwrite(&spell, sizeof(spell), 1, handle);
    fwrite(&power, sizeof(power), 1, handle);
    time.save(handle);
    fwrite(&decreaseTime, sizeof(decreaseTime), 1, handle);
}

//---------------------------------------------------------------------------------------------
void CSpell::load(FILE *handle)
{
    fread(&spell, sizeof(spell), 1, handle);
    fread(&power, sizeof(power), 1, handle);
    time.load(handle);
    fread(&decreaseTime, sizeof(decreaseTime), 1, handle);
}
