#include "tiles.h"
#include <QtXml/qdom.h>
#include <QFile>
#include "../common/sources/system/filecache.h"
#include "../common/sources/system/cgraphics2d.h"
#include "game.h"
#include <QDebug>
#include "player.h"
#include "interface.h"
#include "scripts.h"

CTiles   tiles;

//---------------------------------------------------------------------------------------------
CTiles::CTiles()
{
    pressPlateStates = NULL;
}

//---------------------------------------------------------------------------------------------
void CTiles::initMap()
{
    allocPressPlates();
    for (int i = 0; i < map.mSize.x * map.mSize.y; ++i)
        pressPlateStates[i] = false;
}

//---------------------------------------------------------------------------------------------
CTiles::~CTiles()
{
    freePressPlates();
}

//-----------------------------------------------------------------------------------------
void CTiles::freePressPlates()
{
    if (pressPlateStates != NULL)
        delete[] pressPlateStates;
}

//-----------------------------------------------------------------------------------------
void CTiles::allocPressPlates()
{
    freePressPlates();
    int mapSize = map.mSize.x * map.mSize.y;
    pressPlateStates = new bool[mapSize];
}

//---------------------------------------------------------------------------------------------
void    CTiles::drawFloorOrnate(QImage* image, CVec2 tablePos, int ornate)
{
    COrnateData ornateData = ornatesDatas[ornate];
    bool    flip = false;

    if (tablePos.x >= (int)((WALL_TABLE_WIDTH + 1) / 2))
    {
        tablePos.x = WALL_TABLE_WIDTH - 1 - tablePos.x;
        flip = true;
    }

    COrnateElement  elem = ornateData.table[tablePos.y][tablePos.x];

    if (elem.image.isEmpty() == false)
    {
        QImage  ornateImage = fileCache.getImage(elem.image.toLocal8Bit().constData());
        CVec2   pos = elem.pos;

        if (flip == true)
            pos.x = MAINVIEW_WIDTH - pos.x - ornateImage.width();

        graph2D.drawImage(image, pos, ornateImage, 0, flip);
    }
}

//--------------------------------------------------------------------------------------------------------
void    CTiles::readFloorOrnatesDB()
{
    QDomDocument doc;

    QFile f("databases/floor_ornates.xml");
    f.open(QIODevice::ReadOnly);
    doc.setContent(&f);
    f.close();

    QDomElement root = doc.documentElement();
    QDomElement ornate = root.firstChild().toElement();

    while(!ornate.isNull())
    {
        if (ornate.tagName() == "floor_ornate")
        {
            COrnateData   newOrnate;

            newOrnate.name = ornate.attribute("name");

            QDomElement ornateInfo = ornate.firstChild().toElement();

            while(!ornateInfo.isNull())
            {
                QString tag = ornateInfo.tagName();

                if (tag.startsWith("image_") == true)
                {
                    std::string str = tag.toUtf8().constData();
                    int x = str.at(str.size() - 2) - '0';
                    int y = str.at(str.size() - 1) - '0';
                    COrnateElement* elem = &newOrnate.table[y][x];
                    elem->image = QString("gfx/3DView/floor_ornates/") + ornateInfo.text();
                }
                else if (tag.startsWith("pos_") == true)
                {
                    std::string str = tag.toUtf8().constData();
                    int x = str.at(str.size() - 2) - '0';
                    int y = str.at(str.size() - 1) - '0';
                    COrnateElement* elem = &newOrnate.table[y][x];
                    elem->pos.x = ornateInfo.attribute("x").toInt();
                    elem->pos.y = ornateInfo.attribute("y").toInt();
                }

                ornateInfo = ornateInfo.nextSibling().toElement();
            }
            ornatesDatas.push_back(newOrnate);
        }
        ornate = ornate.nextSibling().toElement();
    }
}

//---------------------------------------------------------------------------------------------
void    CTiles::readTilesDB()
{
    QDomDocument doc;

    QFile f("databases/tiles.xml");
    f.open(QIODevice::ReadOnly);
    doc.setContent(&f);
    f.close();

    QDomElement root = doc.documentElement();
    QDomElement tile = root.firstChild().toElement();

    while(!tile.isNull())
    {
        std::vector<CParamType> paramsTypesList;

        if (tile.tagName() == "tile")
        {
            QDomElement tileInfo = tile.firstChild().toElement();

            while(!tileInfo.isNull())
            {
                if (tileInfo.tagName() == "param")
                {
                    QString paramType = tileInfo.attribute("type");
                    CParamType  newParamType;

                    if (paramType == "ornate")
                    {
                        newParamType.mType = eParamOrnate;
                        newParamType.mName = tileInfo.text();
                        paramsTypesList.push_back(newParamType);
                    }
                    else if (paramType == "int")
                    {
                        newParamType.mType = eParamInt;
                        newParamType.mName = tileInfo.text();
                        paramsTypesList.push_back(newParamType);
                    }
                    else if (paramType == "bool")
                    {
                        newParamType.mType = eParamBool;
                        newParamType.mName = tileInfo.text();
                        paramsTypesList.push_back(newParamType);
                    }
                    else if (paramType == "door")
                    {
                        newParamType.mType = eParamDoor;
                        newParamType.mName = tileInfo.text();
                        paramsTypesList.push_back(newParamType);
                    }
                    else if (paramType == "door_ornate")
                    {
                        newParamType.mType = eParamDoorOrnate;
                        newParamType.mName = tileInfo.text();
                        paramsTypesList.push_back(newParamType);
                    }
                    else if (paramType == "enum")
                    {
                        newParamType.mType = eParamEnum;
                        newParamType.mName = tileInfo.text();
                        paramsTypesList.push_back(newParamType);
                    }
                    else if (paramType == "script")
                    {
                        newParamType.mType = eParamScript;
                        newParamType.mName = tileInfo.text();
                        paramsTypesList.push_back(newParamType);
                    }
                }

                tileInfo = tileInfo.nextSibling().toElement();
            }
            CMap::mTilesParams.push_back(paramsTypesList);
        }
        tile = tile.nextSibling().toElement();
    }
}

//----------------------------------------------------------------------------
// Update pressure plates states and execute their scripts
void CTiles::updatePressPlates()
{
    for (int y = 0; y < map.mSize.y; ++y)
        for (int x = 0; x < map.mSize.x; ++x)
        {
            CVec2   pos(x, y);
            CTile*  tile = map.getTile(pos);

            bool&   state = pressPlateStates[y * map.mSize.x + x];
            bool    lastState = state;
            state = false;

            if (tile->getType() == eTilePressPlate)
            {
                if (tile->getBoolParam("byChampions") == true && interface.isChampionEmpty(0) == false && player.pos == pos)
                {
                    state = true;
                }
                else if (tile->getBoolParam("byObjects") == true)
                {
                    CVec2   pos2 = pos * 2;

                    if (map.findObjectsStack(pos2, eWallSideMax) != NULL ||
                        map.findObjectsStack(pos2 + CVec2(1, 0), eWallSideMax) != NULL ||
                        map.findObjectsStack(pos2 + CVec2(0, 1), eWallSideMax) != NULL ||
                        map.findObjectsStack(pos2 + CVec2(1, 1), eWallSideMax) != NULL)
                    {
                        state = true;
                    }
                }

                // walk on pressure plate
                if (lastState == false && state == true)
                {
                    if (tile->getBoolParam("playSound") == true)
                        sound->play(pos, "sound/Switch.wav");

                    scripts.execute(tile->getScriptParam("onPressed"));
                }

                // leave pressure plate
                if (lastState == true && state == false)
                    scripts.execute(tile->getScriptParam("onReleased"));
            }
        }
}

//----------------------------------------------------------------------------
struct SStairsData
{
    char        file[32];
    uint16_t    x, y;
};

// stairs going down, facing table
static const SStairsData   gDownFacingStairs[WALL_TABLE_HEIGHT][WALL_TABLE_WIDTH] =
{
    {{"", 0, 0}, {"Down_13_F.png", 8, 63}, {"Down_23_F.png", 75, 58}, {"Down_13_F.png", 141, 63}, {"", 0, 0}}, // Row 3
    {{"", 0, 0}, {"Down_12_F.png", 0, 57}, {"Down_22_F.png", 63, 57}, {"Down_12_F.png", 163, 57}, {"", 0, 0}}, // Row 2
    {{"", 0, 0}, {"Down_11_F.png", 0, 51}, {"Down_21_F.png", 35, 50}, {"Down_11_F.png", 192, 51}, {"", 0, 0}}, // Row 1
    {{"", 0, 0}, {"", 0, 0},               {"", 0, 0},                {"", 0, 0},                 {"", 0, 0}}  // Row 0
};

// stairs going up, facing table
static const SStairsData   gUpFacingStairs[WALL_TABLE_HEIGHT][WALL_TABLE_WIDTH] =
{
    {{"", 0, 0}, {"Up_13_F.png", 14, 59}, {"Up_23_F.png", 78, 58}, {"Up_13_F.png", 147, 59}, {"", 0, 0}}, // Row 3
    {{"", 0, 0}, {"Up_12_F.png",  0, 53}, {"Up_22_F.png", 62, 53}, {"Up_12_F.png", 164, 53}, {"", 0, 0}}, // Row 2
    {{"", 0, 0}, {"Up_11_F.png",  0, 42}, {"Up_21_F.png", 32, 42}, {"Up_11_F.png", 192, 42}, {"", 0, 0}}, // Row 1
    {{"", 0, 0}, {"", 0, 0},              {"", 0, 0},              {"", 0, 0},               {"", 0, 0}}  // Row 0
};

// stairs going down, side table
static const SStairsData   gDownSideStairs[WALL_TABLE_HEIGHT][WALL_TABLE_WIDTH] =
{
    {{"", 0, 0}, {"", 0, 0},                      {"", 0, 0}, {"", 0, 0},                       {"", 0, 0}}, // Row 3
    {{"", 0, 0}, {"Side_12_F.png",      60,  89}, {"", 0, 0}, {"Side_12_F.png",      156,  89}, {"", 0, 0}}, // Row 2
    {{"", 0, 0}, {"Down_Side_11_F.png", 32,  95}, {"", 0, 0}, {"Down_Side_11_F.png", 172,  95}, {"", 0, 0}}, // Row 1
    {{"", 0, 0}, {"Side_10_F.png",       0, 106}, {"", 0, 0}, {"Side_10_F.png",      208, 106}, {"", 0, 0}}  // Row 0
};

// stairs going up, side table
static const SStairsData   gUpSideStairs[WALL_TABLE_HEIGHT][WALL_TABLE_WIDTH] =
{
    {{"", 0, 0}, {"", 0, 0},                    {"", 0, 0}, {"", 0, 0},                     {"", 0, 0}}, // Row 3
    {{"", 0, 0}, {"Side_12_F.png",    60,  89}, {"", 0, 0}, {"Side_12_F.png",    156,  89}, {"", 0, 0}}, // Row 2
    {{"", 0, 0}, {"Up_Side_11_F.png", 32,  91}, {"", 0, 0}, {"Up_Side_11_F.png", 172,  91}, {"", 0, 0}}, // Row 1
    {{"", 0, 0}, {"Side_10_F.png",     0, 106}, {"", 0, 0}, {"Side_10_F.png",    208, 106}, {"", 0, 0}}  // Row 0
};

void CTiles::drawStairs(QImage* image, CTile* tile, CVec2 tablePos)
{
    const SStairsData*  data;

    // find the side of the stairs
    EWallSide   side;
    int i;

    for (i = 0; i < 4; ++i)
        if (tile->mWalls[i].getType() == eWallStairs)
            break;
    side = (EWallSide)i;
    bool isDown = (tile->getIntParam("Level") > game.currentLevel);

    if (side != player.getWallSide(eWallSideDown))
    {
        // get the table data
        if (side == player.getWallSide(eWallSideUp))
        {
            // facing stairs
            if (isDown)
                data = &gDownFacingStairs[tablePos.y][tablePos.x];
            else
                data = &gUpFacingStairs[tablePos.y][tablePos.x];
        }
        else
        {
            // side stairs
            if (isDown)
                data = &gDownSideStairs[tablePos.y][tablePos.x];
            else
                data = &gUpSideStairs[tablePos.y][tablePos.x];
        }

        // display
        if (data->file[0] != 0)
        {
            std::string fileName = std::string("gfx/3DView/stairs/") + std::string(data->file);
            CVec2   pos = CVec2(data->x, data->y);
            bool    flip = (tablePos.x > 2);
            QImage  stairsImage = fileCache.getImage(fileName);
            graph2D.drawImage(image, pos, stairsImage, 0, flip);
        }
    }
    else if (tablePos == CVec2(2, 3))
    {
        // special case when we are on the stairs' tile
        if (isDown)
        {
            QImage  stairsImage = fileCache.getImage("gfx/3DView/stairs/Down_20_L.png");
            graph2D.drawImage(image, CVec2(0, 109), stairsImage);
            graph2D.drawImage(image, CVec2(194, 109), stairsImage, 0, true);
        }
        else
        {
            QImage  stairsImage = fileCache.getImage("gfx/3DView/stairs/Up_20_L.png");
            graph2D.drawImage(image, CVec2(0, 91), stairsImage);
            graph2D.drawImage(image, CVec2(194, 91), stairsImage, 0, true);
        }
    }
}

//----------------------------------------------------------------------------
struct STelepTile
{
    char        file[16];
    uint16_t    x, y;
};

static const STelepTile   gTelepTiles[WALL_TABLE_HEIGHT][WALL_TABLE_WIDTH][3] =
{
    // Row 3 (farthest)
    {
        {{"Wall02_F.png",   0, 58}, {"", 0, 0},                {"Wall13_L.png",  8, 58}}, // 02
        {{"Wall12_F.png",   7, 58}, {"", 0, 0},                {"Wall23_L.png", 78, 59}}, // 12
        {{"Wall22_F.png",  77, 58}, {"", 0, 0},                {"", 0, 0}              }, // 22
        {{"Wall32_F.png", 146, 58}, {"Wall23_R.png", 134, 59}, {"", 0, 0}              }, // 32
        {{"Wall42_F.png", 216, 58}, {"Wall33_R.png", 180, 58}, {"", 0, 0}              }  // 42
    },

    // Row 2
    {
        {{"", 0, 0},                {"", 0, 0},                {"Wall12_L.png",  0, 57}}, // 01
        {{"Wall11_F.png",   0, 52}, {"", 0, 0},                {"Wall22_L.png", 60, 52}}, // 11
        {{"Wall21_F.png",  59, 52}, {"", 0, 0},                {"", 0, 0}              }, // 21
        {{"Wall31_F.png", 164, 52}, {"Wall22_R.png", 146, 52}, {"", 0, 0}              }, // 31
        {{"", 0, 0},                {"Wall32_R.png", 216, 57}, {"", 0, 0}              }  // 41
    },

    // Row 1
    {
        {{"", 0, 0},                {"", 0, 0},                {"", 0, 0}              }, // 00
        {{"Wall10_F.png",   0, 42}, {"", 0, 0},                {"Wall21_L.png", 33, 42}}, // 10
        {{"Wall20_F.png",  32, 42}, {"", 0, 0},                {"", 0, 0}              }, // 20
        {{"Wall30_F.png", 191, 42}, {"Wall21_R.png", 164, 42}, {"", 0, 0}              }, // 30
        {{"", 0, 0},                {"", 0, 0},                {"", 0, 0}              }  // 40
    },

    // Row 0 (nearest)
    {
        {{"", 0, 0},                {"", 0, 0},                {"", 0, 0}             }, // 00
        {{"", 0, 0},                {"", 0, 0},                {"Wall20_L.png", 0, 33}}, // 10
        {{"Wall20_F.png",  32, 42}, {"Wall20_R.png", 191, 33}, {"Wall20_L.png", 0, 33}}, // 20
        {{"", 0, 0},                {"Wall20_R.png", 191, 33}, {"", 0, 0}             }, // 30
        {{"", 0, 0},                {"", 0, 0},                {"", 0, 0}             }  // 40
    }
};

void CTiles::drawTeleporter(QImage* image, CVec2 tablePos)
{
    for (int i = 0; i < 3; ++i)
    {
        int random = game.frameCounter / 20;
        const STelepTile* tileInfo = &gTelepTiles[tablePos.y][tablePos.x][i];

        if (tileInfo->file[0] != 0)
        {
            static char fileName[256];
            sprintf(fileName, "gfx/3DView/walls/%s", tileInfo->file);
            QImage  wallGfx = fileCache.getImage(fileName);
            CVec2   wallPos(tileInfo->x, tileInfo->y);
            QImage  teleportGfx = fileCache.getImage("gfx/3DView/Teleporter.png");

            // create the teleporter image
            QImage  tempImg(wallGfx.size(), QImage::Format_ARGB32);
            tempImg.fill(TRANSPARENT);

            int subImgX = (wallGfx.size().width() + teleportGfx.size().width() - 1) / teleportGfx.size().width();
            int subImgY = (wallGfx.size().height() + teleportGfx.size().height() - 1) / teleportGfx.size().height();
            random = (random * 31415821 + 1) % 100000000;

            for (int y = 0; y < subImgY; ++y)
                for (int x = 0; x < subImgX; ++x)
                {
                    CVec2   subImgPos(x * teleportGfx.size().width(),
                                      y * teleportGfx.size().height());
                    bool flip = ((random & 1) != 0);
                    int angle = ((random >> 1) % 4) * 90;
                    graph2D.drawImage(&tempImg, subImgPos, teleportGfx, angle, flip);
                }

            // darken
            int shadow = WALL_TABLE_HEIGHT - 2 - tablePos.y;
            if (shadow < 0)
                shadow = 0;
            graph2D.darken(&tempImg, (float)shadow * 0.2f);

            // draw the image
            graph2D.drawImageWithMask(image, wallPos, tempImg, wallGfx);
        }
    }
}

//----------------------------------------------------------------------------
bool* CTiles::getPressPlates()
{
    return pressPlateStates;
}

//----------------------------------------------------------------------------
void CTiles::setPressPlates(bool* table)
{
    allocPressPlates();
    for (int i = 0; i < map.mSize.x * map.mSize.y; ++i)
        pressPlateStates[i] = table[i];
}
